#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include "struct.h"

#include "io.h"
#include "client.h"
#include "config.h"

char *filebuffer;
struct mtpheader *header;
struct mtpmessage *mtpm;

#define BUFSIZE 2048
char mtphost[BUFSIZE];
int mtpport;

char thisgame[BUFSIZE];

/*------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
   int i;

   thisgame[0] = '\0';
   open_all();

   if (argc == 1)
      print_help();

   for (i = 1; i < argc; i++)
   {
      if (*argv[i] == '-')	/* Argument ist ein Flag */
	 parse_flag(argv[i]);
      else			/* Argument ist ein File */
	 transmit(argv[i]);
   }
   close_all();
}

/*------------------------------------------------------------------------*/

void print_help(void)
{
   printf("USAGE: mtp [-r] [-g<game>] file1 file2 ...\n");
   printf("            -r: Receive file from MUD\n");
   printf("      -g<game>: Specify the target MUD\n");
   close_all();
   exit(2);
}

/*------------------------------------------------------------------------*/

void parse_flag(char *arg)
{
   int i;

   for (i = 1; i < strlen(arg); i++)
      switch (arg[i])
      {
	 case 'r':
	    short2chars(chars2short(header->flags) | RECEIVE, header->flags);
	    break;
	 case 'o':
	    short2chars(chars2short(header->flags) | OVERWRITE, header->flags);
	    break;
	 case 'g':
	    strcpy(thisgame, arg + 2);
	    get_params(thisgame, mtphost, &mtpport,
		       header->player, header->passwd);
	    return;
	 default:
	    print_help();
      }
}

/*------------------------------------------------------------------------*/

void transmit(char *path)
{
   int fh, sh, bytes;
   char tmp[5];

   short2chars(chars2short(header->flags) | VERSION_TWO, header->flags);
      /* Receive File from Host */
   if (chars2short(header->flags) & RECEIVE)
   {
      printf("Trying %s ... please be patient !\n", mtphost);
      get_file(path, header);
      return;
   }
      /* Send File to Host */
   if (check_directory(path))
   {
      printf("%s is a directory (not transmitted) !\n", path);
      return;
   }
      /* Open file on local host */
   if ((fh = open(path,           O_RDONLY, 4)) < 1 &&
       (fh = open(cut_path(path), O_RDONLY, 4)) < 1)
   {
      fprintf(stderr, "Error: File '%s' not found or accessible...\n", path);
      close_all();
      exit(0);
   }
   printf("Trying %s ... please be patient !\n", mtphost);

      /* Read file from local host into buffer */
   if ((bytes = read(fh, filebuffer, MAX_FILELENGTH)) >= MAX_FILELENGTH)
   {
      fprintf(stderr,
              "Error: File too large. Maximum filesize is %d Bytes !\n",
              MAX_FILELENGTH);
      close_all();
      exit(0);
   }
   close(fh);

   short2chars(bytes, header->filelength);

   if (*header->directory != '\0')
   {
      strcpy(header->filename, header->directory);
      cat_path(header->filename, cut_path(path));
   }
   else
      strcpy(header->filename, path);

      /* Connect to MTP Server */
   if (!(sh = connect_socket(mtphost, mtpport)))
   {
      printf("No connection to MTPserver. Transmission failed !\n");
      exit(0);
   }
      /* Send TestByte */
   strcpy(tmp, "@");
   safe_write(sh, tmp, 1);

      /* Send MTPHeader */
   if (!(safe_write(sh, (char *) header, sizeof(struct mtpheader))))
   {
      printf("Couldn't write MTPheader ... aborting\n");
      close_all();
      exit(0);
   }
      /* Receive Message from Server */
   if (!(read_message(sh, mtpm)))
      return;

      /* Send Filebuffer to Server */
   if (!(safe_write(sh, filebuffer, chars2short(header->filelength))))
   {
      printf("\nCouldn't transmit datafile ... aborted.\n");
      close_all();
      exit(0);
   }
      /* Receive Message from Server */
   if (!(read_message(sh, mtpm)))
      return;
}

/*-------------------------------------------------------------------------*/

void open_all(void)
{

   header = (struct mtpheader *) introduce(sizeof(struct mtpheader));

   filebuffer = introduce(MAX_FILELENGTH + 1);
   mtpm = (struct mtpmessage *) introduce(sizeof(struct mtpmessage));

   get_params(thisgame, mtphost, &mtpport, header->player, header->passwd);

}

/*-------------------------------------------------------------------------*/

void close_all(void)
{
   dismiss((char *)header);
   dismiss(filebuffer);
   dismiss((char *)mtpm);
}
