
/*
 * io.c : Enthaelt die Routinen fuer die Sockets und
 * Logfiles
 */

/*-----------------------------------------------------------------------*/

#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#if defined(linux) || defined(solaris)
#	include <unistd.h>
#endif
#include <sys/stat.h>
#if !defined(solaris)
#	include <sys/dir.h>
#endif

#include "config.h"
#include "struct.h"
#include "io.h"


#define BUFSIZE 2048

/*-----------------------------------------------------------------------*/

extern struct mtpheader *header;
extern struct mtpmessage *mtpm;
extern char mtphost[BUFSIZE];
extern int mtpport;
extern char *filebuffer;

/*-----------------------------------------------------------------------*/

int create_socket(int port)

     /*
     * create_socket() oeffnet eine Socket (s), nagelt sie
     * auf der Adressstruktur (sin) fest und wartet
     * auf eine Verbindung (listen). Kommt diese zustande,
     * wird ueber accept() ein Level1-Filehandler
     * zurueckgegeben. Diese Routine ist das mieseste
     * Stueck Patchwork, was ich je gesehen habe, aber sie
     * funktioniert (und sichert sich damit ihre
     * Daseinsberechtigung).
     */

{
   int on = 1;
   static int s = 0;
   static int g, len;
   static struct sockaddr_in sin, from;

   len = sizeof(&from);
   if (!s)
   {
      memset((char *) &sin, 0, sizeof(sin));
      sin.sin_port =        htons(port);
      sin.sin_addr.s_addr = htonl(INADDR_ANY);
      sin.sin_family =      AF_INET;

      if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	 return 0;
#if defined(solaris)
      (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
#endif
      if (bind(s, (struct sockaddr *) &sin, sizeof(sin)) < 0)
      {
	 close(s);
	 return 0;
      }
   }
   listen(s, 5);
   g = accept(s, (struct sockaddr *) &from, &len);
   return g;
}

/*------------------------------------------------------------------------*/

int connect_socket(char *host, int port)

     /*
     * connect_socket() versucht eine Verbindung zum
     * Serverhost aufzustellen (open_socket). Gelingt das,
     * werden die angegebenen Daten uebertragen.
     */

{
   int s;

   return (s = open_socket(host, port)) == -1 ? 0 : s;
}

/*------------------------------------------------------------------------*/

int open_socket(char *h_name, int portnr)

     /*
     * open_socket() versucht eine Verbindung zum
     * Serverhost herzustellen. Gelingt dies, wird die
     * Socketnummer zurueckgegeben, andernfalls -1.
     */

{
   int res;
   int s;
   struct in_addr addr;
   struct sockaddr_in sin;
   struct hostent *host;

   if (!((*h_name >= '0') && (*h_name <= '9')))
      if ((host = gethostbyname(h_name)))
	 h_name = (char *) inet_ntoa(host->h_addr_list[0]);
      else
      {
	 printf("open_socket: Couldn't find host %s .\n", h_name);
	 return 0;
      }
   addr.s_addr = inet_addr(h_name);
   if ((int) addr.s_addr == -1)
   {
      printf("open_socket: Couldn't find host %s .\n", h_name);
      return 0;
   }
   memset((char *) &sin, 0, sizeof(sin));
   sin.sin_family = AF_INET;
   sin.sin_port   = htons(portnr);
   memcpy((char *) &sin.sin_addr, &addr, sizeof(struct sockaddr_in));

   if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      perror("open_socket: socket");
      return -1;
   }
   if (res = connect(s, (struct sockaddr *) &sin, sizeof(sin)) < 0)
   {
      perror("open_socket: connect");
      return -1;
   }
   return s;
}

/*------------------------------------------------------------------------*/

int log_message(struct mtpheader *mtph) 

     /*
     * log_message baut einen String folgender Form auf:
     * Tag Monat Datum hh:mm:ss <Wizard> Filename (# Bytes)
     * Diese Message wird an das MTPLOGFILE appended.
     */

{
   FILE *fd;
   time_t t;
   char timestr[30];

   time(&t);
   strcpy(timestr, ctime(&t));
   timestr[strlen(timestr) - 6] = '\0';		/* Das unnuetze Jahr aus
						 * dem String nehmen */
   if (!(fd = fopen(MTPLOGFILE, "a")))
   {
      printf("Couldn't write to MTPLOGFILE ... check config.h for the correct path !\n");
      close_all();
      exit(0);
   }
   fprintf(fd, "%s <%s> %s: %s (%d Bytes)\n", 
           timestr,
           mtph->player,
           (chars2short(mtph->flags) & RECEIVE) ? "recv" : "send",
           mtph->filename, 
           chars2short(mtph->filelength));
   fclose(fd);
}

/*--------------------------------------------------------------------------*/

char * introduce(int length)

     /*
     * introduce() ist meine Standardinitialisierung fuer
     * Puffer, Structs etc. Es ist nichts als ein leicht
     * geschoentes malloc() mit dem Zusatz, dass der
     * Speicher sicher geleert wird. Sieht auf den ersten
     * Blick umstaendlich aus, erspart aber VIEL Aerger mit
     * Verfuegbarkeitsbereichen etc.
     */

{
   char *pointer;

   pointer = (char *) malloc(length);
   memset(pointer, 0, length);
   return pointer;
}

/*--------------------------------------------------------------------------*/

void dismiss(char *pointer)

     /*
     * dismiss() gibt die mit introduce() initialisierten
     * Variablen wieder frei
     */
{
   free(pointer);
}

/*-------------------------------------------------------------------------*/

int safe_write(int sh, char *buffer, long length)
{
   int bytes, cy;

   for(cy = 0; length > 0; length -= bytes, buffer += bytes)
      if((bytes = write(sh, buffer, length)) > 0)
         cy = 0;
      else if(++cy > 10)
         return 0;
   return 1;
}

/*------------------------------------------------------------------------*/

int safe_read(int sh, char *buffer, long length)
{
   int bytes, cy;

   for(cy = 0; length > 0; length -= bytes, buffer += bytes)
      if((bytes = read(sh, buffer, length)) > 0)
	 cy = 0;
      else if(++cy > 10)
	 return 0;
   return 1;
}

/*-------------------------------------------------------------------------*/

int read_message(int sh, struct mtpmessage *m)
{
   if (!(safe_read(sh, (char *)m, sizeof(struct mtpmessage))))
        return 0;
   printf(m->message);
   return !(m->quit == 1);
}

/*-----------------------------------------------------------------------*/

int write_message(int sh, struct mtpmessage *mtpm, char *message, int quit)
{
   strcpy(mtpm->message, message);
   mtpm->quit = quit;
   return safe_write(sh, (char *)mtpm, sizeof(struct mtpmessage)) > 0;
}

/*-----------------------------------------------------------------------*/

int mtp_open(char *path, struct mtpheader *mtph)
{
   return open(path, O_WRONLY | O_CREAT | O_TRUNC, 0660);
}

/*-----------------------------------------------------------------------*/

char *build_path(struct mtpheader *mtph)
{
   static char path[100];

   strcpy(path, MUDPATH);
   if (chars2short(mtph->flags) & VERSION_TWO)
   {
      if (*mtph->filename != '/')
      {
	 cat_path(path, "w");
	 cat_path(path, mtph->player);
      }
      cat_path(path, mtph->filename);
   }
   else
   {
      cat_path(path, "w");
      cat_path(path, mtph->player);
      if (*mtph->directory != '\0')
	 cat_path(path, cut_path(mtph->directory));
      cat_path(path, cut_path(mtph->filename));
   }
   return path;
}

/*-----------------------------------------------------------------------*/

char *cut_path(char *path)
{
   int i;

   for (i = strlen(path) - 1; i >= 0; i--)
      switch (path[i])
      {
	 case '/':
	 case '~':
	    return path + i + 1;
	 case '.':
	    if (i && path[i - 1] == '.')
	       return path + i + 1;
      }
   return (path);
}

/*-----------------------------------------------------------------------*/

/* Send file from host to client */
void send_file(int sh, char *path, struct mtpheader *mtph, char *mess)
{
   int fh, bytes;
   char tmp[300];

   if ((fh = open(path, O_RDONLY, 4)) < 1)
   {
      strcat(mess, " aborted.\nFile not found or accessible on mud!\n");
      write_message(sh, mtpm, mess, 1);
      return;
   }
   if ((bytes = read(fh, filebuffer, MAX_FILELENGTH)) >= MAX_FILELENGTH)
   {
      sprintf(tmp, "%s aborted.\nFile too large. Maximum filesize is "
                   "%d Bytes!\n", mess, MAX_FILELENGTH);
      write_message(sh, mtpm, tmp, 1);
      return;
   }
   close(fh);

   short2chars(bytes, header->filelength);

   sprintf(tmp, "%s %d bytes", mess, bytes);
   if (!write_message(sh, mtpm, tmp, 0))
      return;
   if (!(safe_write(sh, (char *) header, sizeof(struct mtpheader))))
      return;
   if (!(safe_write(sh, filebuffer, bytes)))
      return;
   log_message(mtph);
}

/*--------------------------------------------------------------------------*/

/* Receive File from Host by client */
void get_file(char *file, struct mtpheader *mtph)
{
   char *path, tmp[200];
   int sh, fh;

   short2chars(0, header->filelength);

   if (!access(file, F_OK))
   {
      printf("Error: Cannot write to a existing file ... aborted\n");
      close_all();
      exit(0);
   }
   if (!(sh = connect_socket(mtphost, mtpport)))
   {
      printf("No connection to MTPserver. Transmission failed !\n");
      close_all();
      exit(0);
   }
      /* Send testbyte to host */
   strcpy(tmp, "@");
   safe_write(sh, tmp, 1);

   strcpy(header->filename, file);

      /* Send MTPheader to host */
   if (!(safe_write(sh, (char *) header, sizeof(struct mtpheader))))
   {
      printf("\nCouldn't write MTPheader ... aborting\n");
      close_all();
      exit(0);
   }
   if (!(read_message(sh, mtpm)))
      return;

      /* Receive MTPHeader from Host */
   if (!(safe_read(sh, (char *)mtph, sizeof(struct mtpheader))))
   {
      printf("\nCouldn't read MTPheader ... aborting\n");
      close_all();
      exit(0);
   }
   if (*mtph->filename == '/')
      path = cut_path(mtph->filename);
   else
      path = mtph->filename;

      /* Open client local file */
   if ((fh = open(path, O_WRONLY | O_CREAT | O_EXCL,0600)) < 1)
   {
      if ((fh = open(cut_path(path), O_WRONLY | O_CREAT | O_EXCL,0600)) < 1)
      {
	 printf("\nCouldn't write destination file (existing ?) "
	        "... Transmission aborted.\n");
	 close_all();
	 exit(0);
      }
      strcpy(path, cut_path(path));
   }
      /* Receive file from host */
   if (!(safe_read(sh, filebuffer, chars2short(mtph->filelength))))
   {
      printf(" terminated.\nError during file transfer ...\n");
      close_all();
      exit(0);
   }
   else
      printf(" completed.\n");

      /* Write File to client disk */
   safe_write(fh, filebuffer, chars2short(mtph->filelength));

   if (fh)
      close(fh);
}

/*-------------------------------------------------------------------------*/

void print_header(struct mtpheader *mtph)
{
   printf("Name : %s\n", mtph->player);
   printf("Passwort : %s\n", mtph->passwd);
   printf("Dateiname : %s\n", mtph->filename);
   printf("Dateilaenge : %d\n", chars2short(mtph->filelength));
   printf("Status : %d\n", chars2short(mtph->flags));
}

/*-------------------------------------------------------------------------*/

void get_params(char *thisgame, char *host, int *port, 
                char *wizard, char*password)
{
   char *hdir;
   char buf[BUFSIZE];
   FILE *cfile;
   char game[BUFSIZE];
   struct stat stb;

   hdir = (char *) getenv("HOME");
   if (hdir == NULL)
      hdir = ".";
   sprintf(buf, "%s/.mtprc", hdir);
   if (!(cfile = fopen(buf, "r")) && 
       !(cfile = fopen(".mtprc", "r")))
   {
      printf("Couldn't read .mtprc ! Adjust it and copy it to your Homedir.\n");
      exit(1);
   }
   if (fstat(fileno(cfile), &stb) >= 0 && (stb.st_mode & 077) != 0)
   {
      printf("Error - .mtprc file not correct mode. "
             "First do chmod 700 ~/.mtprc\n");
      exit(1);
   }
   while (fgets(buf, BUFSIZE, cfile) != NULL)
   {
      if (sscanf(buf, "game %s host %s port %u wizard %s password %s",
		 game, host, port, wizard, password) == 5)
      {
	 if (!strcmp(game, thisgame))
	    return; /* got it */
      }
      else
      if (!thisgame[0] && sscanf(buf, "default %s", thisgame) != 1)
	 printf("Mangled line in .mtprc: %s", buf);
   }
   printf("Game %s not found. Check your .mtprc !\n", thisgame);
   exit(1);
}

/*------------------------------------------------------------------------*/

char * cat_path(char *path, char *add)
{
   char *point;

   point = add;
   if (!strlen(path))
      return strcpy(path, add);
   if ((path[strlen(path) - 1] == '/') && (add[0] == '/'))
      point++;
   if ((path[strlen(path) - 1] != '/') && (add[0] != '/'))
      strcat(path, "/");
   strcat(path, point);
   return (path);
}

/*-----------------------------------------------------------------------*/

/* UNItopia MTP Access Control for Read/Write Operations
 *
 * fullpath    Unixpath to requested file on mudhost
 * path        Mudlib related path to requested file
 * wizard      Name of requesting wizard
 * is_write    1 = write to mudlib
 *             0 = read from mudlib
 *
 * valid_access() retcode: 1 = access granted
 *                         0 = access denied
 */

#define MAX_TOKENS 3
#define L_PATH 200

int valid_access(char *fullpath, char *path, char *wizard, int is_write)
{
   char tokpath[L_PATH], cutfullpath[L_PATH], dir[L_PATH],
     tmppath[L_PATH];
   char *parts[MAX_TOKENS], name[L_PATH];
   char format[10];
   FILE *access_file;
   int i;

   /* Illegal Characters => No Access */
   if (strstr(path, "..") || strchr(path, '~'))
      return 0;
   /* Full Access to own Directories */
   if (*path != '/')
      return 1;

   /* Absolute Path to directory */
   strcpy(dir, fullpath);
   *strrchr(dir, '/') = '\0';

   /* Tokenize mudpath */
   strcpy(tokpath, path);
   for (i = 1, parts[0] = strtok(tokpath, "/");
	i < MAX_TOKENS && (parts[i] = strtok(NULL, "/"));
	i++);
   if (is_write)
   {
      /* WRITE only in
       * own /p or /w directory
       * own Domaindirectory
       * /open directory
       */
      if (parts[1] && (!strcmp(parts[0], "w") || !strcmp(parts[0], "p")) &&
	  !strcmp(parts[1], wizard) ||
	  parts[2] && !strcmp(parts[0], "d") && !strcmp(parts[2], wizard) ||
	  parts[0] && !strcmp(parts[0], "open"))
	 return 1;
   }
   else
   {
      /* READ all, but
       * /var directories
       * other /w/ directories
       *    [ if no READ access file with wizara'ds name or 'alle' in it found ]
       */
      if (parts[0] && !strcmp(parts[0], "var"))
	 return 0;
      if (parts[1] && !strcmp(parts[0], "w") && strcmp(parts[1], wizard))
      {
	 strcpy(tmppath, dir);
	 strcat(tmppath, "/READ");
	 if ((access_file = fopen(tmppath, "r")) != NULL)
	 {
#           if defined(solaris)
	       while (fscanf(access_file, "%199s", name) != EOF)
#           else
	       while (fscanf(access_file, "%*s", L_PATH-1, name) != EOF)
#           endif
		  if (!strcasecmp(name, wizard) || !strcasecmp(name, "alle"))
		     return 1;
	    fclose(access_file);
	 }
	 return 0;
      }
   }

   /* WRITE in not owned Domaindirectories, if
    *       WRITE access file with wizard's name in it there found
    */
   strcpy(tmppath, dir);
   strcat(tmppath, "/WRITE");
   if (is_write && parts[0] && !strcmp(parts[0], "d") &&
       (access_file = fopen(tmppath, "r")) != NULL)
   {
      sprintf(format, "%%%ds", L_PATH-1);
      while (fscanf(access_file, format, name) != EOF)
	 if (!strcasecmp(name, wizard))
	    return 1;
      fclose(access_file);
   }

   /* No MTP access in (nor read nor write) if mtplock lockfile found in
      a mudlib directory */
   strcpy(tmppath, dir);
   strcat(tmppath, "/");
   strcat(tmppath, LOCKFILE);

   return access(tmppath, 0) && !is_write;
}

/*-------------------------------------------------------------------------*/

int check_directory(char *path)
{
   return opendir(path) ? 1 : 0;
}

/*-------------------------------------------------------------------------*/

void short2chars(unsigned short us, unsigned char uc[2])
{
   uc[0] = us / 256;
   uc[1] = us & 255;
}

/*-------------------------------------------------------------------------*/

unsigned short chars2short(unsigned char uc[2])
{
   return (uc[0] * 256) + uc[1];
}
